"""Various utility functions for pop2net."""

from __future__ import annotations

import inspect
import typing

import numpy as np


def group_it(
    value: int | float,
    start: int | float,
    step: int,
    n_steps: int,
    return_value: typing.Literal["index", "lower_bound", "range"] = "index",
    summarize_highest: bool = False,
) -> int | float | tuple[int | float, int | float]:
    """_summary_.

    Args:
        value: _description_
        start: _description_
        step: _description_
        n_steps: _description_
        return_value: _description_. Defaults to "index".
        summarize_highest: _description_. Defaults to False.

    Returns:
        _type_: _description_
    """
    assert type(value) in [int, float], f"{value} has to be a number!"

    if value < start:
        new_value = np.nan

    else:
        for i in range(n_steps):
            lower_bound = start + step * i
            upper_bound = lower_bound + step

            if lower_bound <= value < upper_bound:
                break

        if return_value == "index":
            new_value = i

        elif return_value == "lower_bound":
            new_value = lower_bound

        elif return_value == "range":
            new_value = (lower_bound, upper_bound)

        if value >= upper_bound:
            if not summarize_highest:
                new_value = np.nan

    return new_value


def print_header(text: object):
    """Print a header around an object.

    Args:
        text: Object to be printed within the header.
    """
    # TODO: This is weird.... and it should probably not be print-statements cuz no one can catch
    # them.
    print("")
    print("")
    print("______________________________________")
    print(text)
    print("______________________________________")
    print("")


def _to_list(x: object) -> list:
    return x if isinstance(x, list) else [x]


def _get_cls_as_str(cls_):
    # if isinstance(cls_, str):
    #    # if cls_ is a string
    #    return cls_
    if inspect.isclass(cls_):
        # if cls_ is a class
        return cls_.__name__
    else:
        # if cls_ is an instance
        return cls_.__class__.__name__


def _join_positions(pos1, pos2):
    return "-".join(sorted([str(pos1), str(pos2)]))



"""Base class to create Location objects."""

from __future__ import annotations

from . import actor as _actor


class Location:
    """Base class for location objects."""

    label: str | None = None

    def __init__(self, *args, **kwargs) -> None:
        """Location constructor."""
        self.label = self.__class__.__name__ if self.label is None else self.label
        self.env = None
        self.id_p2n = None
        self.model = None
        self.type = type(self).__name__
        super().__init__(*args, **kwargs)

    def setup(self):
        pass

    @property
    def actors(self) -> list:
        """Return the list of actors affiliated with this location.

        Returns:
            List of actors at this location.
        """
        return self.env.actors_of_location(self)

    def add_actor(self, actor: _actor.Actor, weight: float | None = None) -> None:
        """Assigns the given actor to this location.

        Args:
            actor: The actor that should be added to the location.
            weight: The edge weight between the actor and the location.
                Defaults to None. If weight is set to None, the weight is generated by
                location.weight(), which returns 1 by default.
        """
        self.env.add_actor_to_location(self, actor=actor, weight=weight)

    def add_actors(self, actors: list, weight: float | None = None) -> None:
        """Add multiple actors at once.

        Args:
            actors (list): An iterable over actors.
            weight(float | None): The edge weight between the actors and the location.
                Defaults to None. If weight is set to None, the weight is generated by
                location.weight(), which returns 1 by default.

        """
        for actor in actors:
            self.add_actor(actor=actor, weight=weight)

    def remove_actor(self, actor: _actor.Actor) -> None:
        """Removes the given actor from this location.

        Args:
            actor: Actor that is to be removed.
        """
        self.env.remove_actor_from_location(
            location=self,
            actor=actor,
        )

    def remove_actors(self, actors: list) -> None:
        """Remove multiple actors at once.

        Args:
            actors (list): An iterable over actors.
        """
        for actor in actors:
            self.remove_actor(actor=actor)

    def neighbors(self, actor: _actor.Actor) -> ActorList:
        """Returns a list of actors which are connected to the given actor via this location.

        Args:
            actor: Actor of whom the neighbors at this location are to be returned.

        Returns:
            ActorList: A list of all actors at this location except the passed actor.
        """
        actors = self.env.actors_of_location(self)
        actors.remove(actor)
        return actors

    def set_weight(self, actor, weight: float | None = None) -> None:
        """Set the weight of an actor at the current location.

        If weight is None the method location.weight() will be used to generate a weight.

        Args:
            actor (Actor): The actor.
            weight (float): The weight.
        """
        self.env.set_weight(
            actor=actor,
            location=self,
            weight=weight,
        )

    def get_weight(self, actor: _actor.Actor) -> float:
        """Return the edge weight between an actor and the location.

        Args:
            actor: Actor of which the edge weight should be returned.

        Returns:
            Edge weight.
        """
        return self.env.get_weight(actor=actor, location=self)

    def weight(self, actor) -> float:  # noqa: ARG002
        """Generates the edge weight between a given actor and the location instance.

        Args:
            actor (_type_): An actor.

        Returns:
            float: The weight between the given actor and the location.
        """
        return 1

    def project_weights(self, actor1: _actor.Actor, actor2: _actor.Actor) -> float:
        """Calculates the edge weight between two actors assigned to the same location instance.

        Defines how the weights are combined when the edge weight between two actors is determined.
        Can be completely rewritten to have location-specific methods of this kind with the same
        name or can be used as it is in the simulation code.

        Args:
            actor1: First actor of the pair.
            actor2: Second actor of the pair.

        Raises:
            Exception: Raised if `self.weight_projection_function` is not in ["average", "simultan"]

        Returns:
            Combined edge weight.
        """
        return min([self.get_weight(actor1), self.get_weight(actor2)])



"""Create a population for the simulation."""

from __future__ import annotations

import itertools
import math
import random
import warnings

import pandas as pd

import pop2net as p2n
import pop2net.utils as utils

from .exceptions import Pop2netException


class Creator:
    """Creates and connects actors and locations."""

    def __init__(
        self,
        env: p2n.Environment,
        seed: int = None,
    ) -> None:
        """Instantiate a creator for a specific model.

        Args:
            seed (int, optional): A seed for reproducibility. Defaults to 999.
        """
        self.env = env
        self.seed = seed
        self.rng = random.Random(seed)
        self._temp_actor_attrs = ["_P2NTEMP_split_values", "_P2NTEMP_melt_location_weight"]
        self.model = self.env.model

    def _create_dummy_location(self, designer) -> p2n.Location:
        # TODO: Organize this code better

        if self.env.framework is None:
            if designer.location_class is None:
                # case 1: no framework & default location class
                dummy_location_class = type("Location", (designer,), {})
            else:
                # case 2: no framework & custom location class
                location_name = utils._get_cls_as_str(designer.location_class)
                dummy_location_class = type(location_name, (designer, designer.location_class), {})

            dummy_location = dummy_location_class()
            dummy_location.model = (
                self.model
            )  # if no model was passed to the env, this is just None

        else:
            if designer.location_class is None:
                # case 3: framework & default location class
                dummy_location_class = type("Location", (designer, self.env._framework.Agent), {})
            else:
                # case 4: framework & custom location class
                location_name = utils._get_cls_as_str(designer.location_class)
                dummy_location_class = type(location_name, (designer, designer.location_class), {})
            dummy_location = dummy_location_class(model=self.model)

        dummy_location.label = (
            designer.label if designer.label is not None else utils._get_cls_as_str(designer)
        )
        dummy_location.setup()
        return dummy_location

    def draw_sample(
        self,
        df: pd.DataFrame,
        n: int | None = None,
        sample_level: str | None = None,
        sample_weight: str | None = None,
        replace_sample_level_column: bool = True,
    ) -> pd.DataFrame:
        """Draw a sample from a dataframe.

        Args:
            df (:class:`pandas.DataFrame`): a pandas DataFrame
            n (int): Target size of the final sample. If this is higher the size of the input
                DataFrame, the sampling will occure with replacement. Without otherwise. If `n` is
                set to `None`, df is returned as it is.
            sample_level (str, optional): A variable the specifies sample units,
                i.e. rows that should always
            be sampled together. For instance, a household ID to sample by households.
                Defaults to None.
            sample_weight (str, optional): The column of df which should be used as
                probability weight. Defaults to None.
            replace_sample_level_column (bool): Should the original values of the sample level be
                overwritten by unique values after sampling to avoid duplicates?

        Returns:
            A pandas DataFrame.
        """
        df = df.copy()

        if n is None:
            return df

        elif sample_level is None:
            df_sample = df.sample(
                n=n,
                replace=not (n <= len(df) and sample_weight is None),
                weights=sample_weight,
                random_state=self.seed,
            )

        else:
            sample_level_ids = list(df[sample_level].drop_duplicates())

            if sample_weight is not None:
                weights = list(df.drop_duplicates(subset=sample_level)[sample_weight])

            samples = []
            counter = 0
            sample_cluster_id = 1
            while counter < n:
                if sample_weight is None:
                    random_id = self.rng.choice(sample_level_ids)
                else:
                    random_id = self.rng.choices(sample_level_ids, weights=weights, k=1)[0]

                sample = df.loc[df[sample_level] == random_id, :].copy()

                # create new unique ids for sample level variable
                if replace_sample_level_column:
                    sample.loc[:, sample_level + "_original"] = sample.loc[:, sample_level]
                    sample.loc[:, sample_level] = sample_cluster_id

                samples.append(sample)

                counter += len(sample)
                sample_cluster_id += 1

            df_sample = pd.concat(samples).reset_index(drop=True)

        return df_sample

    def create_actors(
        self,
        actor_class=None,
        actor_class_attr: None | str = None,
        actor_class_dict: None | dict = None,
        df: pd.DataFrame | None = None,
        n: int | None = None,
        clear: bool = False,
    ):
        """Creates actors from a pandas DataFrame.

        Creates one actor-instance of the given actor-class for each row of the given df,
        if df is not None.
        All columns of the df are added as instance attributes containing the row-specific values
        of the specific column.
        If df is None and n is not None, n default actors without any additional attributes are
        created.

        Args:
            actor_class: A class to instantiate all actors with. Every column in the DataFrame will
                result in an attribute of the actors.
            df: The DataFrame from which the actors should be created from.
            n: The number of actors that should be created. Defaults to None.
            clear (bool): Should the actors already included in the model be removed?

        Returns:
            A list of actors.
        """
        if clear:
            self.env.remove_actors(self.env.actors)

        # if no actor class was provided
        if actor_class is None:
            # if no framework ist used, use default actor class
            if self.env.framework is None:
                actor_class = p2n.Actor

            # if a framework is used, create mixed actor class
            else:

                class MixedActor(p2n.Actor, self.env._framework.Agent):
                    pass

                actor_class = MixedActor

        if df is not None:
            df = df.copy()

            # create one actor for each row in df
            actors = []
            for _, row in df.iterrows():
                if actor_class_dict is None:
                    if self.env.framework is None:
                        actor = actor_class()
                        actor.model = (
                            self.model
                        )  # if no model was passed to the env, this is just None
                    else:
                        actor = actor_class(model=self.model)
                else:
                    actor = actor_class_dict[row[actor_class_attr]](model=self.model)

                for col_name in df.columns:
                    if col_name == "id":
                        msg = "You are not allowed to set an actor attribute called `id`."
                        raise Exception(msg)
                    else:
                        setattr(actor, col_name, row[col_name])
                actors.append(actor)

        else:
            if n is not None:
                if self.model is None:
                    actors = [actor_class() for _ in range(n)]
                else:
                    actors = [actor_class(model=self.model) for _ in range(n)]

            else:
                msg = "Either `df` or `n` must be not None."
                raise Exception(msg)

        # add actors to environment
        self.env.add_actors(actors)

        return self.env._to_framework(actors)

    def _get_affiliated_actors(self, actors, dummy_location) -> list:
        temp_filter_attr = "_P2NTEMP_filter_" + dummy_location.label
        self._temp_actor_attrs.append(temp_filter_attr)

        affiliated_actors = []
        for actor in actors:
            if not hasattr(actor, temp_filter_attr):
                setattr(actor, temp_filter_attr, dummy_location.filter(actor))
                self._temp_actor_attrs.append(temp_filter_attr)

            if getattr(actor, temp_filter_attr):
                affiliated_actors.append(actor)

        return affiliated_actors

    def _get_mother_group_id(self, actor, dummy_location) -> str:
        if dummy_location.nest() is None:
            return "None"

        else:
            # search for mother location assigned to this actor
            n_mother_locations_found = 0
            for location in actor.locations:
                if location.label == dummy_location.nest():
                    mother_location = location
                    n_mother_locations_found += 1

            # Check if the number of mother locations is not 1
            if n_mother_locations_found > 1 and self.env.enable_p2n_warnings:
                warnings.warn(
                    f"""For actor {actor},
                    {n_mother_locations_found} locations of class
                    {dummy_location.nest()} were found as potential mothers of
                    {dummy_location} in the list of locations.""",
                    stacklevel=2,
                )
            elif n_mother_locations_found == 0:
                return "None"

            return "-".join([str(mother_location.split_value), str(mother_location.group_id)])

    def _get_split_values(
        self,
        actors: list,
        dummy_location,
        allow_nesting: bool = False,
    ) -> list[int | str]:
        all_values = []
        for actor in actors:
            actor_values = utils._to_list(dummy_location.split(actor))

            if allow_nesting:
                # Add mother location's value to the value of the lower level location
                for i, value in enumerate(actor_values):
                    mother_group_id = self._get_mother_group_id(actor, dummy_location)
                    actor_values[i] = "-".join([mother_group_id, str(value)])
                    temp_attr = f"_P2NTEMP_{dummy_location.label}_mother_group_id"
                    setattr(actor, temp_attr, mother_group_id)
                    self._temp_actor_attrs.append(temp_attr)

            # Temporarely store group values as actor attribute
            # to assign them to the corresponding location group later
            actor._P2NTEMP_split_values = actor_values

            for value in actor_values:
                if value not in all_values:
                    all_values.append(value)

        return all_values

    def _get_stick_value(self, actor, dummy_location):
        stick_value = dummy_location.stick_together(actor)
        if stick_value is None:
            return "None" + str(actor.id_p2n)
        else:
            return stick_value

    def _get_groups(self, actors, designer) -> list[list]:
        overcrowding_i = 0

        dummy_location = self._create_dummy_location(designer)

        n_location_groups_is_fixed = False

        # determine the number of groups needed
        if dummy_location.n_locations is None and dummy_location.n_actors is None:
            n_location_groups = 1
            groups: list[list] = [[]]

        elif dummy_location.n_locations is None and dummy_location.n_actors is not None:
            if dummy_location.overcrowding is None:
                round_function = round
            elif dummy_location.overcrowding is True:
                round_function = math.floor
            elif dummy_location.overcrowding is False:
                round_function = math.ceil
            else:
                # TODO
                raise Exception

            n_location_groups = max(
                round_function(len(actors) / dummy_location.n_actors),
                1,
            )
            groups: list[list] = [[]]

        elif dummy_location.n_locations is not None and dummy_location.n_actors is None:
            n_location_groups = dummy_location.n_locations
            designer._P2NTEMP_ori_n_actors = dummy_location.n_actors
            designer.n_actors = max(
                math.floor(len(actors) / n_location_groups),
                1,
            )

            groups: list[list] = [[] for _ in range(n_location_groups)]

        elif dummy_location.n_locations is not None and dummy_location.n_actors is not None:
            n_location_groups = dummy_location.n_locations
            n_location_groups_is_fixed = True
            groups: list[list] = [[]]

        else:
            # TODO:
            raise Exception

        stick_values = {self._get_stick_value(actor, dummy_location) for actor in actors}

        # dummy_location = self._create_dummy_location(designer)

        # for each group of sticky actors
        for stick_value in stick_values:
            sticky_actors = [
                actor
                for actor in actors
                if self._get_stick_value(actor, dummy_location) == stick_value
            ]

            assigned = False

            for _, group in enumerate(groups):
                # if there are still enough free places available
                if (dummy_location.n_actors is None) or dummy_location.n_actors - len(group) >= len(
                    sticky_actors
                ):
                    # if sum(
                    #    [dummy_location.find(actor) for actor in sticky_actors],
                    # ) == len(sticky_actors):
                    #    # assign actors
                    for actor in sticky_actors:
                        group.append(actor)
                        # dummy_location.add_actor(actor)

                    assigned = True
                    break

            if not assigned:
                if len(groups) < n_location_groups:
                    new_group = []
                    dummy_location = self._create_dummy_location(designer)
                    # assign actors
                    for actor in sticky_actors:
                        new_group.append(actor)
                        # dummy_location.add_actor(actor)

                    groups.append(new_group)

                else:
                    if not dummy_location.only_exact_n_actors and not n_location_groups_is_fixed:
                        for actor in sticky_actors:
                            groups[overcrowding_i].append(actor)
                        overcrowding_i = (overcrowding_i + 1) % len(groups)

        if dummy_location.only_exact_n_actors:
            groups = [group for group in groups if len(group) == dummy_location.n_actors]

        return groups

    def _get_split_value_affiliated_actors(
        self,
        actors: list,
        split_value: int | str,
    ) -> list:
        group_affiliated_actors = [
            actor for actor in actors if split_value in actor._P2NTEMP_split_values
        ]

        return group_affiliated_actors

    def _get_melted_groups(self, actors: list, designer) -> list[list]:
        dummy_location = self._create_dummy_location(designer)

        # get all mother locations the actors are nested in
        all_mother_group_ids = {
            self._get_mother_group_id(actor, dummy_location) for actor in actors
        }

        # for each mother location
        for mother_group_id in all_mother_group_ids:
            # get actors that are part of this location
            nested_actors = [
                actor
                for actor in actors
                if self._get_mother_group_id(actor, dummy_location) == mother_group_id
            ]

            # a list that stores a list of groups for each location
            # [
            # [[_actor], [_actor], [_actor]], # groups of location 1
            # [[_actor], [_actor]],           # groups of location 2
            # ]
            groups_to_melt_by_location: list[list[list]] = []

            # for each location that shall be melted
            for melt_designer in dummy_location.melt():
                # create dummy location
                melt_dummy_location = self._create_dummy_location(melt_designer)
                if melt_designer.n_locations is None:
                    melt_designer.n_locations = designer.n_locations

                # get all actors that should be assigned to this location
                # filter by melt_location
                melt_location_affiliated_actors = self._get_affiliated_actors(
                    actors=nested_actors,
                    dummy_location=melt_dummy_location,
                )
                # filter by main_location
                melt_location_affiliated_actors = self._get_affiliated_actors(
                    actors=melt_location_affiliated_actors,
                    dummy_location=dummy_location,
                )

                # get all values for which separated groups/locations should be created
                melt_split_values = self._get_split_values(
                    actors=melt_location_affiliated_actors,
                    dummy_location=melt_dummy_location,
                    allow_nesting=False,
                )

                for actor in melt_location_affiliated_actors:
                    actor._P2NTEMP_melt_location_weight = melt_dummy_location.weight(actor)
                    self._temp_actor_attrs.append("_P2NTEMP_melt_location_weight")

                # for each split value: get groups and collect them in one list for all values
                location_groups_to_melt: list[list] = []
                for melt_split_value in melt_split_values:
                    melt_split_value_affiliated_actors = self._get_split_value_affiliated_actors(
                        actors=melt_location_affiliated_actors,
                        split_value=melt_split_value,
                    )
                    location_groups_to_melt.extend(
                        self._get_groups(
                            actors=melt_split_value_affiliated_actors,
                            designer=melt_designer,
                        ),
                    )

                groups_to_melt_by_location.append(location_groups_to_melt)

            # Melt groups
            all_melted_groups: list[list] = []
            z = sorted(
                [len(groups_to_melt) for groups_to_melt in groups_to_melt_by_location],
                reverse=True if dummy_location.recycle else False,
            )[0]

            for i in range(z):
                melted_group = []
                for groups_to_melt in groups_to_melt_by_location:
                    if len(groups_to_melt) > 0:
                        if dummy_location.recycle:
                            melted_group.extend(groups_to_melt[i % len(groups_to_melt)])
                        else:
                            try:
                                melted_group.extend(groups_to_melt[i])
                            except IndexError:
                                pass

                all_melted_groups.append(melted_group)

        return all_melted_groups

    def _create_designer_mutations(self, location_designers: list) -> list:
        # for each location designer class
        for designer in location_designers:
            # create a dummy location to use the location's methods
            dummy_location = self._create_dummy_location(designer)

            # if the user defined mutations
            if dummy_location.mutate() is not None:
                mutations = []

                # create all possible combinations of the mutation values
                # {key: [1, 2, 3], key2: [4, 5]} -> [{key: 1, key2: 4}, {key: 1, key2: 5}, ...]
                combinations = [
                    dict(zip(dummy_location.mutate().keys(), values))
                    for values in itertools.product(*dummy_location.mutate().values())
                ]

                # create a new location designer class for each combination
                for i, combo in enumerate(combinations):
                    mutation = type(
                        dummy_location.type,
                        (designer,),
                        {
                            **combo,
                            "label": f"{dummy_location.label}{i}",
                        },
                    )
                    mutations.append(mutation)

                # store the mutations in the designer class
                designer._P2NTEMP_mutations = mutations

        # replace the original location designer classes with the mutations
        location_designers_with_mutations = []
        for location_designer in location_designers:
            if hasattr(location_designer, "_P2NTEMP_mutations"):
                location_designers_with_mutations.extend(location_designer._P2NTEMP_mutations)
            else:
                location_designers_with_mutations.append(location_designer)

        return location_designers_with_mutations

    def create_locations(
        self,
        location_designers: list,
        actors: list | p2n.ActorList | None = None,
        clear: bool = False,
        delete_magic_actor_attributes: bool = True,
    ) -> p2n.LocationList:
        """Creates location instances and connects them with the given actor population.

        Args:
            location_designers (list): A list of LocationDesigner classes.
            actors (list | p2n.ActorList): A list of actors.
            clear (bool): Should the locations already included in the model be removed?
            delete_magic_location_attributes (bool): If True, all magic location attributes will be
                removed after the creation of the location instances.
            delete_magic_actor_attribtues (book): If True, all magic actor attributes will be
                removed after the creation of the location instances.

        Returns:
            p2n.LocationList: A list of locations.
        """
        # Create designer mutations
        location_designers = self._create_designer_mutations(location_designers)

        # Check if all location classes have a unique label
        labels = []
        for designer in location_designers:
            dummy_location = self._create_dummy_location(designer)
            if dummy_location.label is None:
                msg = (
                    f"LocationDesigner class {designer} has no label. Please define a unique label."
                )
                raise Pop2netException(msg)
            elif dummy_location.label in labels:
                msg = f"""LocationDesigner class {designer} has a duplicate label. 
                    Please define a unique label."""
                raise Pop2netException(msg)
            else:
                labels.append(dummy_location.label)

        # Remove all locations if clear is True
        if clear:
            self.env.remove_locations(self.env.locations)

        # Use the existing actors in the model if no actors are given if actors is None:
        if actors is None:
            actors = self.env.actors

        # Create a list containing the names of all special location attributes
        # to delete those attributes later
        magic_actor_attributes = []

        # Create magic actor attributes for each location designer class
        for designer in location_designers:
            dummy_location = self._create_dummy_location(designer)
            label = dummy_location.label
            for actor in actors:
                setattr(actor, label, None)
                magic_actor_attributes.append(label)

                setattr(actor, label + "_assigned", False)
                magic_actor_attributes.append(label + "_assigned")

                setattr(actor, label + "_id", None)
                magic_actor_attributes.append(label + "_id")

                setattr(actor, label + "_position", None)
                magic_actor_attributes.append(label + "_position")

                setattr(actor, label + "_head", None)
                magic_actor_attributes.append(label + "_head")

                setattr(actor, label + "_tail", None)
                magic_actor_attributes.append(label + "_tail")

        # The list of all created location instances
        locations = []

        # For each designer: start of creation procedure
        for designer in location_designers:
            # Set temporary location weight of MeltLocations to None
            for actor in actors:
                actor._P2NTEMP_melt_location_weight = None

            # create location dummy in order to use the location's methods
            dummy_location = self._create_dummy_location(designer)
            label = dummy_location.label

            # If nxgraph is used do some checks
            if dummy_location.nxgraph is not None:
                if dummy_location.n_actors is not None and self.env.enable_p2n_warnings:
                    msg = """You cannot define location.n_actors if location.nxgraph is used. 
                        It will be set to the number of nodes in location.nxgraph automatically."""
                    warnings.warn(msg)
                designer.n_actors = len(list(dummy_location.nxgraph.nodes))
                dummy_location.n_actors = len(list(dummy_location.nxgraph.nodes))

                if dummy_location.overcrowding is True and self.env.enable_p2n_warnings:
                    msg = """You cannot define location.overcrowding if location.nxgraph is used. 
                        It will be set to `False` automatically."""
                    warnings.warn(msg)
                designer.overcrowding = False
                dummy_location.overcrowding = False

            # check if n_locations and split is used at the same time
            if (
                not all(dummy_location.split(actor) is None for actor in actors)
                and dummy_location.n_locations is not None
            ):
                msg = f"""You cannot use {label}.n_locations and {label}.split() at the same time.
                {label}.n_locations is ignored and set to None."""
                warnings.warn(msg)
                designer.n_locations = None
                dummy_location.n_locations = None

            # bridge
            if not dummy_location.melt():
                bridge_values = {
                    dummy_location.bridge(actor)
                    for actor in self._get_affiliated_actors(
                        actors=actors, dummy_location=dummy_location
                    )
                    if dummy_location.bridge(actor) is not None
                }

                if len(bridge_values) == 0:
                    pass

                elif len(bridge_values) == 1 and self.env.enable_p2n_warnings:
                    msg = f"""{dummy_location.label}.bridge() returned only one unique value.
                    {dummy_location.label}.bridge() must return at least two unique values in order 
                    to create locations that bring together actors with different values on the 
                    same attribute.
                    """
                    warnings.warn(msg)

                elif len(bridge_values) > 1 and self.env.enable_p2n_warnings:
                    if dummy_location.n_actors is not None:
                        msg = f"""You cannot use {label}.n_actors and 
                        {label}.bridge() at the same time. {label}.n_actors
                        is ignored."""
                        warnings.warn(msg)

                    melt_list = []

                    # create one MeltLocation for each bridge_value
                    for bridge_value in bridge_values:

                        def filter(self, actor):
                            return dummy_location.bridge(actor) == self.bridge_value

                        dummy_melt_class = type(
                            f"dummy_meltlocation{str(bridge_value)}",
                            (p2n.MeltLocationDesigner,),
                            {
                                "filter": filter,
                                "n_actors": 1,
                                "bridge_value": bridge_value,
                            },
                        )

                        melt_list.append(dummy_melt_class)

                    # set the created MeltLocations as return values of melt()
                    def melt(self):
                        return melt_list

                    designer.melt = melt
                    dummy_location = self._create_dummy_location(designer)

            if not dummy_location.melt():
                # get all actors that could be assigned to locations of this class
                affiliated_actors = self._get_affiliated_actors(
                    actors=actors,
                    dummy_location=dummy_location,
                )

            else:
                affiliated_actors = []

                for melt_designer in dummy_location.melt():
                    melt_dummy_location = self._create_dummy_location(melt_designer)
                    affiliated_actors.extend(
                        self._get_affiliated_actors(
                            actors=actors,
                            dummy_location=melt_dummy_location,
                        ),
                    )

            # get all values that are used to split the actors into groups
            split_values = self._get_split_values(
                actors=affiliated_actors,
                dummy_location=dummy_location,
                allow_nesting=True,
            )

            if len(split_values) == 0:
                split_values.append("dummy_split_value")

            group_count = 0

            # for each group split value
            for split_value in split_values:
                split_value_locations = []

                # get all actors with that value
                split_value_affiliated_actors = self._get_split_value_affiliated_actors(
                    actors=affiliated_actors,
                    split_value=split_value,
                )

                # if this location does not glue together other locations
                if not dummy_location.melt():
                    group_lists: list[list] = self._get_groups(
                        actors=split_value_affiliated_actors,
                        designer=designer,
                    )
                else:
                    group_lists = self._get_melted_groups(
                        actors=split_value_affiliated_actors,
                        designer=designer,
                    )

                # for each group of actors
                for i, group_list in enumerate(group_lists):
                    group_count += 1

                    dummy_location = self._create_dummy_location(designer)
                    dummy_location.actors_ = group_list

                    # get all subgroub values
                    subsplit_values = {
                        actor_subsplit_value
                        for actor in group_list
                        for actor_subsplit_value in utils._to_list(
                            dummy_location._subsplit(actor),
                        )
                    }

                    # for each group of actors assigned to a specific sublocation
                    for j, subsplit_value in enumerate(subsplit_values):
                        # get all subsplit affiliated actors
                        subsplit_affiliated_actors = []

                        # for actor in group_affiliated_actors:
                        for actor in group_list:
                            actor_subsplit_value = utils._to_list(
                                dummy_location._subsplit(actor),
                            )
                            if subsplit_value in actor_subsplit_value:
                                subsplit_affiliated_actors.append(actor)

                        # inspect the defined magic location class get all methods/attributes
                        # that are not part of magic location class
                        keep_attrs = {}
                        for attr in dir(designer):
                            if attr not in p2n.LocationDesigner.__dict__:
                                keep_attrs[attr] = getattr(designer, attr)

                        # Create the final location class
                        if designer.location_class is None:
                            if self.env.framework is None:
                                location_class = type(
                                    "Location",
                                    (p2n.Location,),
                                    keep_attrs,
                                )

                            else:
                                location_class = type(
                                    "Location",
                                    (
                                        p2n.Location,
                                        self.env._framework.Agent,
                                    ),  # inherit from framework.Actor
                                    keep_attrs,
                                )

                        else:
                            location_class = type(
                                utils._get_cls_as_str(designer.location_class),
                                (designer.location_class,),
                                keep_attrs,
                            )

                        # Create the final location instance
                        if self.env.framework is None:
                            location = location_class()
                        else:
                            location = location_class(model=self.env.model)

                        location.label = (
                            designer.label
                            if designer.label is not None
                            else utils._get_cls_as_str(designer)
                        )
                        location.setup()
                        location.split_value = split_value
                        location.subsplit_value = subsplit_value
                        location.group_id = i
                        location.subgroup_id = j

                        self.env.add_location(location=location)

                        split_value_locations.append(location)

                        # Assigning process:
                        # The if-statement below is just a quick fix to handle .n_actors == 0.
                        # TODO: Fix this earlier in the code
                        if dummy_location.n_actors is None or dummy_location.n_actors > 0:
                            for actor in subsplit_affiliated_actors:
                                location.add_actor(actor)

                                weight = (
                                    actor._P2NTEMP_melt_location_weight
                                    if actor._P2NTEMP_melt_location_weight is not None
                                    else location.weight(actor)
                                )

                                location.set_weight(
                                    actor=actor,
                                    weight=weight,
                                )

                                group_info_str = (
                                    f"gv={location.split_value},gid={location.group_id}"
                                )
                                setattr(actor, label, group_info_str)
                                setattr(actor, label + "_assigned", True)
                                setattr(actor, label + "_id", group_count - 1)
                                setattr(actor, label + "_position", group_list.index(actor))
                                setattr(
                                    actor,
                                    label + "_head",
                                    True if group_list.index(actor) == 0 else False,
                                )
                                setattr(
                                    actor,
                                    label + "_tail",
                                    True
                                    if group_list.index(actor) == (len(group_list) - 1)
                                    else False,
                                )

                            locations.append(location)

                if (
                    dummy_location.n_locations is not None
                    and not dummy_location.only_exact_n_actors
                ):
                    if len(split_value_locations) < dummy_location.n_locations:
                        for _ in range(
                            int(dummy_location.n_locations - len(split_value_locations))
                        ):
                            location = designer()
                            location.setup()
                            location.split_value = split_value
                            location.subsplit_value = None
                            location.group_id = None
                            location.subgroup_id = None

                            self.env.add_location(location=location)
                            locations.append(location)

        # Reset location_designer.n_actors if it was changed
        for designer in location_designers:
            if hasattr(designer, "_P2NTEMP_ori_n_actors"):
                designer.n_actors = designer._P2NTEMP_ori_n_actors
                delattr(designer, "_P2NTEMP_ori_n_actors")

        # delete temporary actor attributes
        for actor in actors:
            for attr in self._temp_actor_attrs:
                if hasattr(actor, attr):
                    delattr(actor, attr)

        magic_actor_attributes = set(magic_actor_attributes)
        if delete_magic_actor_attributes:
            for attr in magic_actor_attributes:
                for actor in actors:
                    delattr(actor, attr)

        return self.env._to_framework(locations)

    def create(
        self,
        df: pd.DataFrame,
        location_designers: list,
        actor_class: type[p2n.Actor] = p2n.Actor,
        actor_class_attr: None | str = None,
        actor_class_dict: None | dict = None,
        n_actors: int | None = None,
        sample_level: str | None = None,
        sample_weight: str | None = None,
        replace_sample_level_column: bool = True,
        clear: bool = False,
        delete_magic_actor_attributes: bool = True,
        return_elements=False,
    ) -> tuple:
        """Creates actors and locations based on a given dataset.

        Combines the Creator-methods `draw_sample()`, `create_actors()` and `create_locations()`.

        Args:
            df (pd.DataFrame): A data set with individual data that forms the basis for
                the creation of actors. Each row is (potentially) translated into one actor.
                Each column is translated into one actor attribute.
            actor_class (type[p2n.Actor]): The class from which the actor instances are created.
            location_designers (list): A list of classes from which the location instances are
                created.
            n_actors (Optional[int], optional): The number of actors that will be created.
                If `n_actors` is set to None, each row of `df` is translated into exactly one actor.
                Otherwise, rows are randomly drawn (with replacement,
                if `n_actors > len(df)`) from `df` until the number of actors created
                equals `n_actors`.
            sample_level (Optional[str], optional): If `sample_level` is None,
                the rows are sampled individually.
                Otherwise the rows are sampled as groups. `sample_level` defines
                which column of `df` contains the group id.
            sample_weight (Optional[str]): The column of df in which should be used as probability
                weight during sampling.
            replace_sample_level_column (bool): Should the original values of the sample level be
                overwritten by unique values after sampling to avoid duplicates?
            clear (bool): Should the actors and locations already included in the model be removed?
            delete_magic_location_attributes (bool): If True, all magic location attributes will be
                removed after the creation of the location instances.
            delete_magic_actor_attributes (bool): If True, all magic actor attributes will be
                removed after the creation of the location instances.

        Returns:
            tuple: A list of actors and a list of locations.
        """
        # draw a sample from dataset
        df_sample = self.draw_sample(
            df=df,
            n=n_actors,
            sample_level=sample_level,
            sample_weight=sample_weight,
            replace_sample_level_column=replace_sample_level_column,
        )

        # create actors
        actors = self.create_actors(
            df=df_sample,
            actor_class=actor_class,
            actor_class_attr=actor_class_attr,
            actor_class_dict=actor_class_dict,
            clear=clear,
        )

        # create locations
        locations = self.create_locations(
            actors=actors,
            location_designers=location_designers,
            clear=clear,
            delete_magic_actor_attributes=delete_magic_actor_attributes,
        )

        if return_elements:
            return actors, locations

    def get_df_actors(
        self,
        columns: None | list[str] = None,
        drop_agentpy_columns: bool = True,
    ) -> pd.DataFrame:
        """Returns the latest created population of actors as a dataframe.

        Args:
            columns (list | None): A list of column names that sould be kept.
                All other columns are deleted.
            drop_agentpy_columns (bool): Deletes some columns created by agentpy.

        Raises:
            Pop2netException: _description_

        Returns:
            pd.DataFrame: A dataframe which contains one row for each
            actor and one column for each actor attribute.
        """
        if self.actors is None:
            msg = "There are no actors."
            raise Pop2netException(msg)

        df = pd.DataFrame([vars(actor) for actor in self.actors])

        if drop_agentpy_columns:
            df = df.drop(
                columns=[
                    "_var_ignore",
                    "id",
                    "type",
                    "log",
                    "model",
                    "p",
                ],
            )

        if columns is not None:
            df = df.loc[:, columns]

        return df



"""A module that helps with creating insights into the network."""

from __future__ import annotations

import typing

from bokehgraph import BokehGraph
import networkx as nx
import pandas as pd
import seaborn as sns
from tabulate import tabulate

import pop2net.utils as utils

if typing.TYPE_CHECKING:
    pass


class NetworkInspector:
    """Helper class that contains functions to inspect the network of a pop2net environment."""

    def __init__(self, env) -> None:
        """Initiate a NetworkInspector.

        Args:
            env (Environment): The environment this NetworkInspector belongs to.
        """
        self.env = env  # TODO: make instances of inspector without a fixed env possible

    def plot_bipartite_network(
        self,
        actor_attrs: list | None = None,
        location_attrs: list | None = None,
        actor_color: str | None = None,
        location_color: str | None = None,
        edge_alpha: str = "weight",
        edge_color: str = "black",
        edge_size: int = 1,
        node_size: int = 10,
        node_alpha=0.5,
    ) -> None:
        """Plots the bipartite network of actors and locations.

        Args:
            actor_attrs (list | None, optional): A list of actor attributes that
                should be shown as node attributes in the network graph. Defaults to None.
            location_attrs (list | None, optional): A list of location attributes that
                should be shown as node attributes in the network graph. Defaults to None.
            actor_color (str | None, optional): The actor attribute that determines the
            color of the actor nodes. Defaults to None.
            location_color (str | None, optional): The location attribute that determines the
            color of the location nodes. Defaults to None.
            edge_alpha (str, optional): The edge attribute that determines the edges' transparency.
                Defaults to "weight".
            edge_color (str, optional): The color of the edges. Defaults to "black".
            edge_size (int, optional): The size of the edges. Defaults to 1.
            node_size (int, optional): The size of the nodes. Defaults to 10.
            node_alpha (float, optional): The transparency of the nodes. Defaults to 0.5.
        """
        if actor_attrs is None:
            actor_attrs = ["type"]
        else:
            actor_attrs = list(actor_attrs)
            if "type" not in actor_attrs:
                actor_attrs.append("type")

        if actor_color is not None and actor_color not in actor_attrs:
            actor_attrs.append(actor_color)

        if location_attrs is None:
            location_attrs = ["type", "label"]
        else:
            location_attrs = list(location_attrs)
            if "type" not in location_attrs:
                location_attrs.append("type")

            if "label" not in location_attrs:
                location_attrs.append("label")

        if location_color is not None and location_color not in location_attrs:
            location_attrs.append(location_color)

        graph = self.env.export_bipartite_network(
            actor_attrs=actor_attrs,
            location_attrs=location_attrs,
        )

        graph_layout = nx.drawing.spring_layout(graph)
        plot = BokehGraph(
            graph,
            width=400,
            height=400,
            hover_edges=True,
            bipartite=True,
        )
        plot.layout(layout=graph_layout)

        plot.draw(
            node_color=(
                actor_color if actor_color is not None else "firebrick",
                location_color if location_color is not None else "steelblue",
            ),
            edge_alpha=edge_alpha,
            edge_color=edge_color,
            edge_size=edge_size,
            node_size=node_size,
            node_alpha=node_alpha,
        )

    def plot_actor_network(
        self,
        actor_color: str | None = None,
        actor_attrs: list | None = None,
        edge_alpha: str = "weight",
        edge_color: str = "black",
        include_0_weights: bool = True,
        edge_size: int = 1,
        node_size: int = 10,
        node_alpha=0.5,
    ) -> None:
        """Plots the actor network.

        Args:
            actor_color (str | None, optional): The actor attribute that determines the
            color of the actor nodes. Defaults to None.
            actor_attrs (list | None, optional): A list of actor attributes that
                should be shown as node attributes in the network graph. Defaults to None.
            edge_alpha (str, optional): The edge attribute that determines the edges' transparency.
                Defaults to "weight".
            edge_color (str, optional): The color of the edges. Defaults to "black".
            include_0_weights (bool, optional): Should edges with a weight of zero be included in
                the plot? Defaults to True.
            edge_size (int, optional): The size of the edges. Defaults to 1.
            node_size (int, optional): The size of the nodes. Defaults to 10.
            node_alpha (float, optional): The transparency of the nodes. Defaults to 0.5.
        """
        if actor_attrs is None:
            actor_attrs = ["type"]
        else:
            actor_attrs = list(actor_attrs)
            if "type" not in actor_attrs:
                actor_attrs.append("type")

        if actor_color is not None and actor_color not in actor_attrs:
            actor_attrs.append(actor_color)

        graph = self.env.export_actor_network(
            node_attrs=actor_attrs,
            include_0_weights=include_0_weights,
        )

        graph_layout = nx.drawing.spring_layout(graph)
        plot = BokehGraph(graph, width=400, height=400, hover_edges=True)
        plot.layout(layout=graph_layout)
        plot.draw(
            node_color="firebrick" if actor_color is None else actor_color,
            node_size=node_size,
            edge_alpha=edge_alpha,
            edge_color=edge_color,
            edge_size=edge_size,
            node_alpha=node_alpha,
        )

    def plot_networks(
        self,
        actor_attrs: list | None = None,
        location_attrs: list | None = None,
        actor_color: str | None = None,
        location_color: str | None = None,
        edge_alpha: str = "weight",
        edge_color: str = "black",
        include_0_weights: bool = True,
        edge_size: int = 1,
        node_size: int = 10,
        node_alpha: float = 0.5,
    ) -> None:
        """Plots the bipartite network and the actor network.

        Args:
            actor_attrs (list | None, optional): A list of actor attributes that
                should be shown as node attributes in the network graph. Defaults to None.
            location_attrs (list | None, optional): A list of location attributes that
                should be shown as node attributes in the network graph. Defaults to None.
            actor_color (str | None, optional): The actor attribute that determines the
                color of the actor nodes. Defaults to None.
            location_color (str | None, optional): The location attribute that determines the
                color of the location nodes. Defaults to None.
            edge_alpha (str, optional): The edge attribute that determines the edges' transparency.
                Defaults to "weight".
            edge_color (str, optional): The color of the edges. Defaults to "black".
            include_0_weights (bool, optional): Should edges with a weight of zero be included in
                the plot? Defaults to True.
            edge_size (int, optional): The size of the edges. Defaults to 1.
            node_size (int, optional): The size of the nodes. Defaults to 10.
            node_alpha (float, optional): The transparency of the nodes. Defaults to 0.5.
        """
        self.plot_bipartite_network(
            actor_color=actor_color,
            actor_attrs=actor_attrs,
            location_color=location_color,
            location_attrs=location_attrs,
            edge_color=edge_color,
            edge_alpha=edge_alpha,
            edge_size=edge_size,
            node_size=node_size,
            node_alpha=node_alpha,
        )

        self.plot_actor_network(
            actor_color=actor_color,
            actor_attrs=actor_attrs,
            edge_color=edge_color,
            edge_alpha=edge_alpha,
            include_0_weights=include_0_weights,
            edge_size=edge_size,
            node_size=node_size,
            node_alpha=node_alpha,
        )

    def eval_affiliations(self, return_data=False) -> tuple[pd.DataFrame, pd.DataFrame] | None:
        """Prints information on the distribution of actors per location and locations per actor.

        Raises:
            Pop2netException: _description_
            Pop2netException: _description_
        """
        df1 = pd.DataFrame(
            [
                {
                    "location_label": location.label,
                    "n_actors": len(location.actors),
                }
                for location in self.env.locations
            ],
        )

        utils.print_header("Number of locations")
        print(df1.location_label.value_counts().to_frame())

        df1 = df1.groupby("location_label").describe()
        df1.columns = df1.columns.droplevel()
        df1 = df1.drop("count", axis=1)

        utils.print_header("Number of actors per location")
        print(df1)

        df2 = pd.DataFrame(
            [
                {
                    "actor_id": actor.id_p2n,
                    "n_affiliated_locations": len(actor.locations),
                }
                for actor in self.env.actors
            ],
        )
        df2 = df2.n_affiliated_locations.describe()
        df2 = df2.drop("count", axis=0)

        utils.print_header("Number of affiliated locations per actor")
        print(df2.to_frame())

        if return_data:
            return df1, df2
        return None

    # TODO: calculate relative freqs
    def create_contact_matrix(
        self,
        actors: list | None = None,
        attr: str = "id",
        weighted: bool = False,
        plot: bool = True,
        annot: bool = False,
        return_df: bool = False,
    ) -> pd.DataFrame:
        """Create a contact matrix as a DataFrame from a given env's actor list.

        Args:
            actors: A list of actors.
            attr: The actor attribute which is shown in the matrix.
            weighted: Should the contacts be weighted? Defaults to False.
            plot: Should the matrix be plotted? Defaults to False.
            annot: Should the plottet matrix be annotated? Defaults to False.
            return_df: Should the data be returned as pandas.DataFrame?

        Returns:
            A DataFrame containing a contact matrix based on `attr`.
        """
        if actors is None:
            actors = self.env.actors

        contact_data = []
        attr_values = []
        pairs = []

        attr_u_name = f"{attr}_u"
        attr_v_name = f"{attr}_v"

        for actor_u in actors:
            attr_u = getattr(actor_u, attr)
            if attr_u is not None:
                attr_values.append(attr_u)

                for actor_v in actor_u.neighbors():
                    attr_v = getattr(actor_v, attr)
                    if attr_v is not None:
                        attr_values.append(attr_v)

                        pair = {actor_u.id_p2n, actor_v.id_p2n}

                        if pair not in pairs:
                            contact_data.append(
                                {
                                    "id_u": actor_u.id_p2n,
                                    attr_u_name: attr_u,
                                    "id_v": actor_v.id_p2n,
                                    attr_v_name: attr_v,
                                    "weight": actor_u.get_actor_weight(actor_v),
                                },
                            )
                            pairs.append(pair)

        attr_values = list(set(attr_values))

        df = pd.DataFrame(index=sorted(attr_values, reverse=True), columns=sorted(attr_values))
        df = df.fillna(0)

        weight_total = 0
        for contact in contact_data:
            weight = contact["weight"] if weighted else 1
            weight_total += weight

            df.loc[contact[attr_u_name], contact[attr_v_name]] = (
                df.loc[contact[attr_u_name], contact[attr_v_name]] + weight
            )
            df.loc[contact[attr_v_name], contact[attr_u_name]] = (
                df.loc[contact[attr_v_name], contact[attr_u_name]] + weight
            )

        if plot:
            g = sns.heatmap(df, annot=annot, vmin=0, fmt="g")
            g.set(xlabel=attr, ylabel=attr)

        if return_df:
            return df

    def network_measures(self, actor_attrs: list[str] = None, weighted: bool = True) -> list[dict]:
        """Calculates common network measures for the actor-level network graph.

        If the created network consist of independent groups of nodes
        subgraphs are created and measures are calculated for each subgraph

        Args:
            actor_attrs: A list of actor attributes
            weighted: Should the network be considered weighted or unweighted

        Return:
            list of dictionaries of the common network measure results
        """
        nx_graph = self.env.export_actor_network(node_attrs=actor_attrs)

        # make distinction between multiple independent networks and one network

        def get_network_measures(nx_graph) -> dict:
            result_dict = {}
            result_dict["n_nodes"] = nx.number_of_nodes(nx_graph)
            result_dict["diameter"] = nx.diameter(nx_graph, weight="weight")
            result_dict["density"] = nx.density(nx_graph)
            result_dict["transitivity"] = nx.transitivity(nx_graph)
            result_dict["avg_clustering"] = nx.average_clustering(
                nx_graph,
                weight="weight",
            )
            result_dict["avg_path_length"] = nx.average_shortest_path_length(
                nx_graph,
                weight="weight",
            )
            return result_dict

        def get_network_measures_no_weight(nx_graph) -> dict:
            result_dict = {}
            result_dict["n_nodes"] = nx.number_of_nodes(nx_graph)
            result_dict["diameter"] = nx.diameter(nx_graph, weight=None)
            result_dict["density"] = nx.density(nx_graph)
            result_dict["transitivity"] = nx.transitivity(nx_graph)
            result_dict["avg_clustering"] = nx.average_clustering(
                nx_graph,
            )
            result_dict["avg_path_length"] = nx.average_shortest_path_length(
                nx_graph,
            )
            return result_dict

        network_components = list(nx.connected_components(nx_graph))
        network_components = sorted(network_components, key=len, reverse=True)

        # create graph for each component and calculate network measures
        result_list = []
        for component in network_components:
            try:
                nx_subgraph = nx_graph.subgraph(component)

            except nx.NetworkXError:
                print("Cant make graph out of component")
                break
            if weighted:
                result_list.append(get_network_measures(nx_subgraph))
            else:
                result_list.append(get_network_measures_no_weight(nx_subgraph))
        return result_list

    def location_crosstab(
        self,
        location_labels: str | list[str],
        actor_attributes: str | list[str],
        output_format="table",
    ) -> list[pd.DataFrame] | None:
        """Generates a crosstabulation of actor attributes for specified location labels.

        Args:
            location_labels (str | list[str]): Location label(s) to filter locations for the crosstab.
            actor_attributes (str | list[str]): Actor attribute(s) to use for the crosstab.
            output_format (str, optional): Output format, either "table" for printed tables or "df" for DataFrame output. Defaults to "table".

        Returns:
            list[pd.DataFrame] | None: List of DataFrames with crosstab results if output_format is "df", otherwise None.
        """
        # Make every Parameter a list
        if location_labels:
            if not isinstance(location_labels, list):
                location_labels = [location_labels]

        if actor_attributes:
            if not isinstance(actor_attributes, list):
                actor_attributes = [actor_attributes]

        # determine eligible locations classes
        valid_locations = []
        if location_labels:
            for location_instance in self.env.locations:
                for locationtype in location_labels:
                    if location_instance.label == locationtype:
                        valid_locations.append(location_instance)
        else:
            valid_locations = list(self.env.locations)

        # create actor df per location instance
        if output_format == "table":
            actor_dfs = {}
            if actor_attributes:
                for i, location_instance in enumerate(valid_locations):
                    title = f"{i + 1}.Location: {str(location_instance.label).split(' ')[0]}"
                    df = pd.DataFrame([vars(actor) for actor in location_instance.actors])
                    df = df[list(actor_attributes)]
                    actor_dfs[title] = df

            for title, df in actor_dfs.items():
                for actor_attribute in actor_attributes:
                    crosstab_table = pd.crosstab(
                        index=df[actor_attribute],
                        columns="count",
                    )

                    print(f"{title}")
                    print(
                        tabulate(
                            crosstab_table,
                            headers=[actor_attribute, "count"],
                            tablefmt="fancy_grid",
                        ),
                    )
                    print("\n")
        if output_format == "df":
            actor_dfs = {}
            result_list = []
            for i, location_instance in enumerate(valid_locations):
                title = f"{i + 1}.Location: {str(location_instance.label).split(' ')[0]}"
                location_type = str(location_instance.label).split(" ")[0]
                df = pd.DataFrame([vars(actor) for actor in location_instance.actors])

                # only keep wanted columns (actor attributes)
                df = df[list(actor_attributes)]

                # add location type as str
                df["location_type"] = location_type

                actor_dfs[title] = df

            for actor_attribute in actor_attributes:
                location_id = 0
                df_list_of_attribute = []
                for _, df in actor_dfs.items():
                    crosstab_table = pd.crosstab(
                        index=df[actor_attribute],
                        columns="count",
                    )
                    crosstab_table.insert(
                        0,
                        "location_id",
                        [location_id] * len(crosstab_table.index),
                    )
                    crosstab_table["location_type"] = df["location_type"][0]

                    df_list_of_attribute.append(crosstab_table)

                    location_id += 1
                df_attribute = pd.concat(df_list_of_attribute, ignore_index=False)

                # fix columns after concat
                df_attribute.reset_index(inplace=True)
                df_attribute.columns.names = ["index"]
                location_array = df_attribute["location_id"].copy()
                df_attribute.drop(columns="location_id", inplace=True)
                df_attribute.insert(0, "location_id", location_array)

                result_list.append(df_attribute)
            return result_list
        return None

    def location_information(
        self,
        location_labels: str | list[str],
        actor_attributes: str | None | list[str] | None = None,
        output_format: str = "table",
    ) -> None | pd.DataFrame:
        """Provides detailed information about actors assigned to specific location instances.

        This method allows filtering by location labels and actor attributes, and can output
        the information either as a formatted table (printed to stdout) or as a pandas DataFrame.

        Args:
            location_labels (str | list[str]): One or more location labels to filter the locations.
            actor_attributes (str | None | list[str], optional): One or more actor attributes to include in the output.
                If None, a default subset of attributes is used. Defaults to None.
            output_format (str, optional): Determines the output format. Use "table" to print a formatted table,
                or "df" to return a pandas DataFrame. Defaults to "table".

        Returns:
            None | pd.DataFrame: Returns None if output_format is "table" (prints to stdout).
                Returns a pandas DataFrame if output_format is "df".
        """
        if location_labels:
            if not isinstance(location_labels, list):
                location_labels = [location_labels]

        if actor_attributes:
            if not isinstance(actor_attributes, list):
                actor_attributes = [actor_attributes]

        # determine eligible locations classes
        valid_locations = []
        if location_labels:
            for location_instance in self.env.locations:
                for locationtype in location_labels:
                    if location_instance.label == locationtype:
                        valid_locations.append(location_instance)
        else:
            valid_locations = list(self.env.locations)

        # create actor df per location instance
        actor_dfs = {}
        if actor_attributes:
            for i, location_instance in enumerate(valid_locations):
                # Create the title of printout
                title = f"{i + 1}.Location: {str(location_instance.label).split(' ')[0]}"
                location_type = str(location_instance.label).split(" ")[0]
                # get all actors per location instance, subset df by actor-attributes
                df = pd.DataFrame([vars(actor) for actor in location_instance.actors])
                df = df[list(actor_attributes)]
                df["location_type"] = location_type
                actor_dfs[title] = df
        else:
            for i, location_instance in enumerate(valid_locations):
                title = f"{i + 1}.Location: {str(location_instance.label).split(' ')[0]}"
                location_type = str(location_instance.label).split(" ")[0]
                df = pd.DataFrame([vars(actor) for actor in location_instance.actors])
                df.drop(df.iloc[:, 0:7], axis=1, inplace=True)
                df["location_type"] = location_type
                actor_dfs[title] = df

        if output_format == "table":
            #### Print Part "Basic"
            for title, df in actor_dfs.items():
                print(f"{title} \n")
                print(tabulate(df, headers="keys", tablefmt="fancy_grid"))
                print("\n")

        if output_format == "df":
            location_id_counter = 0
            df_list = []
            for _, df in actor_dfs.items():
                df.insert(0, "location_id", [location_id_counter] * len(df.index))
                location_id_counter += 1
                df_list.append(df)
            return pd.concat(df_list, ignore_index=True)
        return None



"""Pop2net. An extension package for location-based simulations based on AgentPy."""

from .actor import Actor
from .creator import Creator
from .environment import Environment
from .exceptions import Pop2netException
from .inspector import NetworkInspector
from .location import Location
from .location_designer import LocationDesigner
from .location_designer import MeltLocationDesigner

# from .sequences import LocationList

__all__ = [
    "Actor",
    "Pop2netException",
    "Location",
    "LocationDesigner",
    "MeltLocationDesigner",
    "NetworkInspector",
    "Environment",
    "LocationList",
    "Creator",
]
